package cs451.net;

import cs451.net.event.*;

import cs451.net.handler.*;
import cs451.parser.Host;
import cs451.parser.Parser;
import cs451.tools.Logger;

import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.*;
import java.util.function.BiConsumer;



public abstract class NetManager {
    private static final Integer THREAD_COUNT = 8;

    private static boolean isStopped = false;

    private static final Map<Class<? extends NetEventHandler>, NetEventHandler> nm_listeners = new HashMap<>();

    private static final ThreadPoolExecutor ex = new ThreadPoolExecutor(THREAD_COUNT,THREAD_COUNT,60,TimeUnit.SECONDS,new LinkedBlockingQueue<>());

    private static BiConsumer<NetEventType,NetEvent> onCompleteHandler;
    private static BiConsumer<NetEventType,Exception> onErrorHandler;


    private static void registerNetHandler(NetEventHandler handler){
            nm_listeners.put(handler.getClass(),handler);
    }

    public static void start(Host h, Parser p, BiConsumer<NetEventType,NetEvent> och, BiConsumer<NetEventType,Exception> oeh) {
        onCompleteHandler = och;
        onErrorHandler = oeh;

        registerNetHandler(new NetHandlerSCKT(   NetHandlerSL.class,   NetHandlerDFLT.class));
        registerNetHandler(new NetHandlerSL(     NetHandlerPL.class,   NetHandlerSCKT.class));
        registerNetHandler(new NetHandlerPL(     NetHandlerBEB.class,  NetHandlerSL.class));
        registerNetHandler(new NetHandlerBEB(    NetHandlerURB.class,  NetHandlerPL.class));
        registerNetHandler(new NetHandlerURB(    NetHandlerFIFO.class, NetHandlerBEB.class));
        registerNetHandler(new NetHandlerFIFO(   NetHandlerTOPL.class, NetHandlerURB.class));
        registerNetHandler(new NetHandlerTOPL(   NetHandlerDFLT.class, NetHandlerFIFO.class));
        registerNetHandler(new NetHandlerDFLT(   NetHandlerDFLT.class, NetHandlerDFLT.class));

        nm_listeners.values().forEach(neh -> neh.start(h,p));
        ex.prestartAllCoreThreads();
    }

    public static void stop() {
        isStopped = true;
        nm_listeners.values().forEach(NetEventHandler::stop);
        ex.shutdown();
        System.out.println("NetManager handled "+ex.getCompletedTaskCount()+" tasks during this run.");
    }

    public static boolean isDone() {
        return isStopped || nm_listeners.values().stream().map(NetEventHandler::isDone).reduce(ex.getActiveCount()==0 && ex.getQueue().size()==0,(b1, b2) -> b1 && b2);
    }


//=================================================================================================================
//=================================================================================================================

    public static void handle(Class<? extends NetEventHandler> layer, NetEventType event, NetEvent ne){
        try {
            ex.getQueue().add(new NetEventRunner(ne, layer, event));
        }catch(IllegalStateException ise){
            error(event,ise);
        }
    }

    public static void complete(NetEventType t,NetEvent ne){
        onCompleteHandler.accept(t,ne);
    }

    public static void error(NetEventType t,Exception e){
        onErrorHandler.accept(t,e);
    }

//=================================================================================================================
//=================================================================================================================

    public static void deliver(Class<? extends NetEventHandler> layer, NetEvent ne) {
        handle(layer,NetEventType.DLVR,ne);
    }

    public static void send(Class<? extends NetEventHandler> layer, NetEvent ne) {
        handle(layer,NetEventType.SEND,ne);
    }

    public static void crash(Class<? extends NetEventHandler> layer, NetEvent ne) {
        handle(layer,NetEventType.CRSH,ne);
    }

    public static void send() {
        new NetEventRunner(NetEvent.EMPTY(),NetHandlerTOPL.class,NetEventType.SEND).run();
    }

    public static void deliverSync(Class<? extends NetEventHandler> layer, NetEvent ne) {
        new NetEventRunner(ne,layer,NetEventType.DLVR).run();
    }
    public static void sendSync(Class<? extends NetEventHandler> layer, NetEvent ne) {
        new NetEventRunner(ne,layer,NetEventType.SEND).run();
    }

//=================================================================================================================
//=================================================================================================================

    private static class NetEventRunner implements Runnable {

        private final NetEvent ne;
        private final Class<? extends NetEventHandler> lt;
        private final NetEventType net;

        private NetEventRunner(NetEvent ne, Class<? extends NetEventHandler> lt, NetEventType net) {
            this.ne = ne;
            this.lt = lt;
            this.net = net;
        }

        @Override
        public void run() {
            Logger.debug(this.lt.getSimpleName()+"_"+this.net+ ": " + ((ne.peer==null)?"?":ne.peer.getId()) + " - "+ne.message.toString()+"");

            NetEventHandler nl = nm_listeners.getOrDefault(this.lt, null);
            if (nl != null) {
                nl.onEvent(this.net,ne);
            }else{
                error(this.net,new UnsupportedOperationException("No Handler for "+this.lt));
            }
        }
    }

}


